/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/  
#include <Wire.h>  // Only needed for Arduino 1.6.5 and earlier
#include "SSD1306.h" // alias for `#include "SSD1306Wire.h"`
#include "display1306.h"
#include "sensor18b20.h"
#include "tempLoggerESP32.h"
#include "myNTPClient.h"

//OLED pins to ESP32 GPIOs via this connection:
//OLED_SDA -- GPIO5s
//OLED_SCL -- GPIO4
//OLED_RST -- GPIO16

SSD1306  display(0x3c, 5, 4); // (uint8_t _address, uint8_t _sda, uint8_t _scl) {
OLEDDisplayUi ui ( &display );

//////////////////////////////////////////////////////////////////////
String display1306::m_TextString;
sensorOledData * display1306::m_oledData;
int display1306::m_actPage = 0;
int display1306::m_maxPage = 0;
const uint8_t * display1306::m_FontName = ArialMT_Plain_24;
int display1306::m_FontLines = 8;
String display1306::m_SensorFound;

//////////////////////////////////////////////////////////////////////
display1306& display1306::GetInstance()
{
  static display1306 instance;
  return instance;
}

//////////////////////////////////////////////////////////////////////
void display1306::RunInstance(void * pvParameters)
{
  display1306& instance = GetInstance();

  instance.setup();

  while (1)
  {
    instance.loop();
  }
}

//////////////////////////////////////////////////////////////////////
display1306::display1306()
{
  // create array with 10 display pages / callback functions
  FrameCallback cb[] = {
    display1306::drawFrameSensor0, display1306::drawFrameSensor1, display1306::drawFrameSensor2, display1306::drawFrameSensor3, display1306::drawFrameSensor4,
    display1306::drawFrameSensor5, display1306::drawFrameSensor6, display1306::drawFrameSensor7, display1306::drawFrameSensor8, display1306::drawFrameSensor9
  };

  // This array keeps function pointers to all frames
  // frames are the single views that slide in
  framesText[0] = display1306::drawFrameText; // first page is text page
  for (int i = 0; i < 10; i++)
  {
    framesSensor[i] = cb[i]; // init sensor pages
  }

  m_NewState = DISPLAY_TEXT;    // only text on initialising
  m_State = DISPLAY_UNDEF;      // set to undef
  m_bInitDone = false;
  m_TextEndTime = 0;
  m_MessagePrio = PRIO_UNDEF;

  m_xMessageSemaphore = xSemaphoreCreateMutex();
}

//////////////////////////////////////////////////////////////////////
void display1306::setup()
{
  MyBasecamp::getBasecampSemaphore();

  String descFont = String("Set font size in oled display, valid values are 10, 16, 24.");
  String keyFont = String("font_size");

  Basecamp& iot = MyBasecamp::GetInstance();
  iot.web.addInterfaceElement(keyFont, "input", descFont, "#configform", keyFont);

  // set defaults keyFont
  if (iot.configuration.get(keyFont).length() == 0 )
  {
    iot.configuration.set(keyFont, String("16"));
  }
  int actFontSettings = iot.configuration.get(keyFont).toInt();
  switch (actFontSettings)
  {
    case 10:
    default:
      m_FontName = ArialMT_Plain_10;
      m_FontLines = 8;
      break;

    case 16:
      m_FontName = ArialMT_Plain_16;
      m_FontLines = 3;
      break;

    case 24:
      m_FontName = ArialMT_Plain_24;
      m_FontLines = 2;
      break;
  };

  MyBasecamp::releaseBasecampSemaphore();

  //////////////////////////////////////////////////////////////////////
  pinMode(16, OUTPUT);
  digitalWrite(16, LOW);    // set GPIO16 low to reset OLED
  delay(50);
  digitalWrite(16, HIGH); // while OLED is running, must set GPIO16 in high

  // The ESP is capable of rendering 60fps in 80Mhz mode
  // but that won't give you much time for anything else
  // run it in 160Mhz mode or just set it to 30 fps
  ui.setTargetFPS(30);

  // Customize the active and inactive symbol
  ui.setActiveSymbol(emptySymbol);
  ui.setInactiveSymbol(emptySymbol);

  // You can change this to
  // TOP, LEFT, BOTTOM, RIGHT
  //ui.setIndicatorPosition(BOTTOM);

  // Defines where the first frame is located in the bar.
  //ui.setIndicatorDirection(LEFT_RIGHT);

  // You can change the transition that is used
  // SLIDE_LEFT, SLIDE_RIGHT, SLIDE_UP, SLIDE_DOWN
  ui.setFrameAnimation(SLIDE_UP);
  ui.setTimePerFrame(3000);
  ui.setTimePerTransition(2000);

  // Add frames
  ui.setFrames(framesText, 1);
  m_maxPage = 1;

  // Add overlays
  //ui.setOverlays(overlays, overlaysCount);

  // Initialising the UI will init the display too.
  ui.init();

  display.flipScreenVertically();
}

//////////////////////////////////////////////////////////////////////
void display1306::loop()
{
  int remainingTimeBudget = ui.update();
  if (remainingTimeBudget <= 0)
  {
    return;
  }
  
  // You can do some work here
  // Don't do stuff if you are below your
  // time budget.
  delay(remainingTimeBudget);
  
  // wait for semaphore
  if (xSemaphoreTake(m_xMessageSemaphore, remainingTimeBudget/portTICK_PERIOD_MS) == pdTRUE )
  {
    // active message ?
    if (m_TextEndTime > 0)
    {
      // remove message
      if (millis() >= m_TextEndTime)
      {
        m_MessagePrio = PRIO_UNDEF;
        m_TextEndTime = 0;
        m_TextString = String("");
      }
    }
    m_NewState = (m_TextEndTime > 0) ? DISPLAY_TEXT : DISPLAY_RUN;
  
    // release semaphore
    xSemaphoreGive( m_xMessageSemaphore );
  }
  
  // get formated data from sensor
  m_oledData = sensor18b20::GetInstance().getDisplayText();
  int frames = ((m_oledData->m_nLines-1) / m_FontLines) + 1; //
  
  // maximal 9 temperatur pages and 1 info page
  if(frames > 9)
  {
    frames = 9;
  }

  // changes?
  if((frames != m_maxPage) || (m_State != m_NewState))
  {
    m_State = m_NewState;
    m_maxPage = frames;

    if(m_State == DISPLAY_RUN)
    {
      ui.setFrames(framesSensor, frames);
    }
    else
    {
      ui.setFrames(framesText, 1);
    }
  }
}

//////////////////////////////////////////////////////////////////////
void display1306::addMessage(String msgText, messagePriority_t msgPrio)
{
  // wait for semaphore
  if (xSemaphoreTake(m_xMessageSemaphore, portMAX_DELAY) == pdTRUE )
  {
    bool bSetNewMessage = false;
    unsigned long endTime = 0;

    // set new message? / prio depending
    switch(msgPrio)
    {
      case PRIO_FATAL:
      case PRIO_ERROR:
        endTime = UINT32_MAX; 
        bSetNewMessage = (msgPrio > m_MessagePrio);
        break;
      case PRIO_INFO:
        endTime = millis() + 10000; 
        bSetNewMessage = (msgPrio >= m_MessagePrio);
        break;
      case PRIO_DEBUG:
        endTime = millis() + 5000; 
        bSetNewMessage = (msgPrio >= m_MessagePrio);
        break;
    };

    
    if(bSetNewMessage)
    {
      m_MessagePrio = msgPrio;
      m_TextString = msgText;
      m_TextEndTime = endTime; 
    }
  
    //Serial.println("Release Semaphore AddMessage");
    xSemaphoreGive( m_xMessageSemaphore );
  }
}

//////////////////////////////////////////////////////////////////////
void display1306::initDone()
{
  //Serial.printf("initDone time %d, size=%d\n", m_TextEndTime, m_messageMap.size());

  if (xSemaphoreTake(m_xMessageSemaphore, portMAX_DELAY) == pdTRUE )
  {
    m_bInitDone = true;

    sensor18b20& instanceSensor = sensor18b20::GetInstance();
    int cntSensor = instanceSensor.getNumberSensorCnt();
    if(cntSensor == 1)
    {
      m_SensorFound = String("Id: ") + String(instanceSensor.getSensorCRC16Id(0));
    }
    else
    {
      m_SensorFound = String(cntSensor) + String(" sensors");
    }
    
    if(cntSensor == 0)
    {
       addMessage(String("No sensor found\n"), PRIO_ERROR);
    }
 
    xSemaphoreGive( m_xMessageSemaphore );   
  }
}

//////////////////////////////////////////////////////////////////////
void display1306::drawTopLine(OLEDDisplay *display, OLEDDisplayUiState* state, int16_t x, int16_t y)
{
  static char mBuffer[10];

  display->setFont(ArialMT_Plain_10);

  // left sensor count
  display->setTextAlignment(TEXT_ALIGN_LEFT);
  display->drawString(0 + x, 0 + y, display1306::m_SensorFound);

  if(MyNTPClient::GetInstance().isTimeSynchronized())
  {
    // right time
    DateTime dt(MyNTPClient::GetInstance().getEpochTime());
    snprintf(mBuffer, sizeof(mBuffer), "%02d:%02d:%02d", dt.hour(), dt.minute(), dt.second());

    display->setTextAlignment(TEXT_ALIGN_RIGHT);
    display->drawString(128 + x, 0 + y, String(mBuffer));
  } 
}

//////////////////////////////////////////////////////////////////////
void display1306::drawFrameText(OLEDDisplay *display, OLEDDisplayUiState* state, int16_t x, int16_t y)
{
  m_actPage = 1;
  drawTopLine(display, state, x, y);
  // Demonstrates the 3 included default sizes. The fonts come from SSD1306Fonts.h file
  // Besides the default fonts there will be a program to convert TrueType fonts into this format
  display->setTextAlignment(TEXT_ALIGN_LEFT);
  display->setFont(ArialMT_Plain_10);
  display->drawString(0 + x, 10 + y, m_TextString);
}

//////////////////////////////////////////////////////////////////////
void display1306::drawFrameSensor(OLEDDisplay *display, OLEDDisplayUiState* state, int16_t x, int16_t y, int16_t nDisplay)
{
  m_actPage = nDisplay;

  drawTopLine(display, state, x, y);

  if (m_oledData)
  {
    display->setTextAlignment(TEXT_ALIGN_RIGHT);
    display->setFont(m_FontName);

    switch (m_FontLines)
    {
      case 8: // 10
        display->drawString(128 + x, 12 + y, m_oledData->m_LineString[nDisplay * 8 + 4]);
        display->drawString(128 + x, 24 + y, m_oledData->m_LineString[nDisplay * 8 + 5]);
        display->drawString(128 + x, 36 + y, m_oledData->m_LineString[nDisplay * 8 + 6]);
        display->drawString(128 + x, 48 + y, m_oledData->m_LineString[nDisplay * 8 + 7]);

        display->drawString(64 + x, 12 + y, m_oledData->m_LineString[nDisplay * 8 + 0]);
        display->drawString(64 + x, 24 + y, m_oledData->m_LineString[nDisplay * 8 + 1]);
        display->drawString(64 + x, 36 + y, m_oledData->m_LineString[nDisplay * 8 + 2]);
        display->drawString(64 + x, 48 + y, m_oledData->m_LineString[nDisplay * 8 + 3]);
        break;

      case 3: // 16
        display->drawString(128 + x, 12 + y, m_oledData->m_LineString[nDisplay * 3 + 0]);
        display->drawString(128 + x, 28 + y, m_oledData->m_LineString[nDisplay * 3 + 1]);
        display->drawString(128 + x, 44 + y, m_oledData->m_LineString[nDisplay * 3 + 2]);
        break;

      case 2: // 24
        display->drawString(128 + x, 14 + y, m_oledData->m_LineString[nDisplay * 2 + 0]);
        display->drawString(128 + x, 38 + y, m_oledData->m_LineString[nDisplay * 2 + 1]);
        break;
    };
  }
}

//////////////////////////////////////////////////////////////////////
