/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.

  Based on:
  // OneWire DS18S20, DS18B20, DS1822 Temperature Example
  //
  // http://www.pjrc.com/teensy/td_libs_OneWire.html
  //
  // The DallasTemperature library can do all this work for you!
  // http://milesburton.com/Dallas_Temperature_Control_Library   
****************************************************************************************/  
#include "OneWire.h"
#include "sensor18b20.h"
#include "tempLoggerESP32.h"
#include "Crc16.h"
#include "myNTPClient.h"

//#define SHOW_DALLAS_ERROR        // uncomment to show Dallas ( CRC ) errors on Serial.
#define ONEWIRE_PIN           25   // OneWire Dallas sensors are connected to this pin

OneWire  ds( ONEWIRE_PIN );        // (a 4.7K pull-up resistor is necessary)

//////////////////////////////////////////////////////////////////////
SensorData::SensorData()
{
}

//////////////////////////////////////////////////////////////////////
void SensorData::Configure()
{
  // set addr and crc16 name
  char AddrString[16];
  snprintf(AddrString, sizeof(AddrString), "%02x%02x%02x%02x%02x%02x%02x%02x", m_Addr[0], m_Addr[1], m_Addr[2], m_Addr[3], m_Addr[4], m_Addr[5], m_Addr[6], m_Addr[7]);
  
  Basecamp& iot = MyBasecamp::GetInstance();

  //////////////////////////////////////////////
  String desc = String("Name for Sensor '") + m_crc16Name + String("'.");//String("' , DS18B20 ID: '") + String(AddrString) + String("'.");
  String key = String("Temp_") + String(AddrString);
  iot.web.addInterfaceElement(key, "input", desc, "#configform", key);

  // set defaults m_Name
  if (iot.configuration.get(key).length() == 0 ) 
  {
    String value = String("Temp") + m_crc16Name;
    iot.configuration.set(key, value);
  }
  m_Name = iot.configuration.get(key);
  m_CountGesamt = 0.0;
  m_CountCRC1 = 0.0;
  m_CountCRC2 = 0.0;
  m_CountSensorError = 0.0;
}

//////////////////////////////////////////////////////////////////////
void SensorData::debugPrint()
{
  Serial.printf("%s (%s) = %.2f\n", &m_Name[0], &m_crc16Name[0], m_tempCelcius);
  Serial.printf("%.1f, %.1f, %.1f, %.1f\n", m_CountGesamt, m_CountCRC1, m_CountCRC2, m_CountSensorError);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
sensor18b20& sensor18b20::GetInstance()
{
  static sensor18b20 instance;
  return instance;
}

//////////////////////////////////////////////////////////////////////
void sensor18b20::RunInstance(void * pvParameters)
{
  sensor18b20& instance = GetInstance();
  
  instance.setup();
  instance.m_bSetupFinished = true;
  instance.loop();
}

//////////////////////////////////////////////////////////////////////
sensor18b20::sensor18b20()
{
  m_numberOfFoundSensors = 0;
  m_bSetupFinished = false;
  m_MeasureIndex = 0;
  m_ProcessIndex = 0;
  m_ValidStartTime = "Unknown";
}

//////////////////////////////////////////////////////////////////////
void sensor18b20::setup()
{
  m_numberOfFoundSensors = 0;
  byte currentAddr[8];

  //for(int g=0;g<5;g++)
  {
    
  m_numberOfFoundSensors = 0;

  // improve to find all sensors after reset 
  vTaskDelay( 1*1000 / portTICK_PERIOD_MS); 
  
  ds.reset_search();
  vTaskDelay( 500 / portTICK_PERIOD_MS); 
  while ( ds.search( currentAddr ) && m_numberOfFoundSensors < MAX_NUMBER_OF_SENSORS )
  {
    // set m_Addr
    memcpy(m_Sensor[m_numberOfFoundSensors].m_Addr, currentAddr, sizeof(currentAddr));
    m_numberOfFoundSensors++;
  }
  Serial.printf( "%i Dallas sensors found.\n", m_numberOfFoundSensors );
  
   //   vTaskDelay( 3*1000 / portTICK_PERIOD_MS); 

  }

  
  if ( m_numberOfFoundSensors )
  {
    // set resolution to 12 bit
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      /////////////
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x4E );  // Write Scratchpad
      ds.write( 0x4b );  // high alarm temp, dummy value
      ds.write( 0x46 );  // low alarm temp, dummy value
      // DS1820 and DS18S20 have no configuration register
      if (m_Sensor[thisSensor].m_Addr[0] != DS18S20MODEL) 
      {
         ds.write( 0x7F );  // 12 bit
      }
      
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x48, false );  // COPYSCRATCH, parasite
      delay(20);  // <--- added 20ms delay to allow 10ms long EEPROM write operation (as specified by datasheet)
      ds.reset();
    }

    //////////////////////////////////////////////
    // create default order 
    String defaultOrder = "";
    Crc16 crc;            // build CRC16  
    char crc16Buffer[5];  // build CRC16  
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      snprintf( crc16Buffer, sizeof(crc16Buffer), "%04X", crc.XModemCrc(m_Sensor[thisSensor].m_Addr, 0, 8));    
      m_Sensor[thisSensor].m_crc16Name = String(crc16Buffer);      
      defaultOrder += m_Sensor[thisSensor].m_crc16Name + String(",");    
    }

    Basecamp& iot = MyBasecamp::GetInstance();
    MyBasecamp::getBasecampSemaphore();

    //////////////////////////////////////////////
    // get/set oled_order
    String descOled = String("Set sensor order and visibility in oled display, default all: ") + defaultOrder;
    String keyOled = String("oled_order");
    iot.web.addInterfaceElement(keyOled, "input", descOled, "#configform", keyOled);
  
    // set defaults keyOled
    if (iot.configuration.get(keyOled).length() == 0 ) 
    {
      iot.configuration.set(keyOled, defaultOrder);
    }
    String actOledOrder = iot.configuration.get(keyOled);

    //////////////////////////////////////////////
    // get/set web_order
    String descWeb = String("Set sensor order and visibility in web page, default all: ") + defaultOrder;
    String keyWeb = String("web_order");
    iot.web.addInterfaceElement(keyWeb, "input", descWeb, "#configform", keyWeb);
  
    // set defaults keyOled
    if (iot.configuration.get(keyWeb).length() == 0 ) 
    {
      iot.configuration.set(keyWeb, defaultOrder);
    }
    String actWebOrder = iot.configuration.get(keyWeb);

    //////////////////////////////////////////////////
    // sensoren konfigurieren
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      m_Sensor[thisSensor].Configure();
      m_Sensor[thisSensor].debugPrint();
    }
    MyBasecamp::releaseBasecampSemaphore();

    //////////////////////////////////////////////////
    // fill order map m_SensorOrderOled
    char * pch;
    int actIndex = 0;
  
    pch = strtok (&actOledOrder[0],",");
    while (pch != NULL)
    {
      for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
      {
        if(m_Sensor[thisSensor].m_crc16Name == String(pch))
        {
          m_SensorOrderOled[actIndex] = &m_Sensor[thisSensor];
          //Serial.printf("map:  %d -> %s\n", actIndex, pch);
          break;
        }
      }     
      pch = strtok (NULL, ",");
      actIndex++;
    }

    //////////////////////////////////////////////////
    // fill order map m_SensorOrderWeb
    actIndex = 0;
  
    pch = strtok (&actWebOrder[0],",");
    while (pch != NULL)
    {
      for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
      {
        if(m_Sensor[thisSensor].m_crc16Name == String(pch))
        {
          m_SensorOrderWeb[actIndex] = &m_Sensor[thisSensor];
          Serial.printf("map:  %d -> %s, %d, %d, 0x%x\n", actIndex, pch, m_SensorOrderWeb.size(), thisSensor, &m_Sensor[thisSensor]);
          break;
        }
      }     
      pch = strtok (NULL, ",");
      actIndex++;
    }
  }
}

//////////////////////////////////////////////////////////////////////
void sensor18b20::loop()
{
  /* main temptask loop */
  while (1)
  {
    byte crcRetry = 0;
    unsigned long nowTime = MyNTPClient::GetInstance().getEpochTime();
    unsigned long nextLoop = ((nowTime / 10)+1)*10; // auf nächste 10 Sekunden runden  
    int diffTime = abs(nextLoop-nowTime);
    //Serial.printf("diffTime = %d, nextLoop=%d, nowTime=%d\n", diffTime, nextLoop, nowTime);
    
    if(diffTime == 10) // startup 
    {
      diffTime = 0;
    }
    else if(diffTime > 7)
    {
      diffTime = 7;
    }
 
    vTaskDelay( diffTime*1000 / portTICK_PERIOD_MS);
    
    // trigger measurement
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x44, 0);        // start conversion, with parasite power off at the end
    }

    //wait for conversion ready
    vTaskDelay( 3*1000 / portTICK_PERIOD_MS); 

    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      byte data[12];
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0xBE );         // Read Scratchpad

      //Serial.print( "Sensor " );Serial.print( thisSensor ); Serial.print("  Data = ");
      //Serial.println( present, HEX );
      //Serial.print(" ");
      for ( byte i = 0; i < 9; i++)
      { // we need 9 bytes
        data[i] = ds.read(  );
        //Serial.print(data[i], HEX);
        //Serial.print(" ");
      }
      //Serial.println();

      byte type_s;
      // the first ROM byte indicates which chip
      switch ( m_Sensor[thisSensor].m_Addr[0] )
      {
        case 0x10:
          //Serial.println("  Chip = DS18S20");  // or old DS1820
          type_s = 1;
          break;
        case 0x28:
          //Serial.println("  Chip = DS18B20");
          type_s = 0;
          break;
        case 0x22:
          //Serial.println("  Chip = DS1822");
          type_s = 0;
          break;
        default:
//#ifdef SHOW_DALLAS_ERROR
          Serial.println("Device is not a DS18x20 family device.");
//#endif
          //return;
          m_Sensor[thisSensor].m_CountSensorError += 1.0;
          continue;
      }

      int16_t raw;
      if ( OneWire::crc8(data, 8) != data[8])
      {
#ifdef SHOW_DALLAS_ERROR
        // CRC of temperature reading indicates an error, so we print a error message and discard this reading
        Serial.print( millis() / 1000.0 ); Serial.print( " - CRC error from device " ); Serial.println( thisSensor );
#endif
        if(crcRetry < 2)
        {
          if(crcRetry == 0)
          {
            m_Sensor[thisSensor].m_CountCRC1 += 1.0;
          }
          else
          {
            m_Sensor[thisSensor].m_CountCRC2 += 1.0;
          }
          crcRetry++;
          thisSensor--;          
          continue; // retry with this sensor
        }

      }
      else
      {
        raw = (data[1] << 8) | data[0];
        if (type_s)
        {
          raw = raw << 3; // 9 bit resolution default
          if (data[7] == 0x10)
          {
            // "count remain" gives full 12 bit resolution
            raw = (raw & 0xFFF0) + 12 - data[6];
          }
        }
        else
        {
          byte cfg = (data[4] & 0x60);
          // at lower res, the low bits are undefined, so let's zero them
          if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
          else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
          else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
          //// default is 12 bit resolution, 750 ms conversion time
        }
        m_Sensor[thisSensor].m_tempCelcius = raw / 16.0;
      } // end if/else valid data
      m_Sensor[thisSensor].m_CountGesamt += 1.0;
    } // end for
    //Serial.printf("measure %.2f\n", m_Sensor[0].m_tempCelcius);

    m_MeasureIndex++; // measure loop finished

    // loop 255  
    if(m_MeasureIndex == (byte)-1)
    {
      m_MeasureIndex = 1;
      m_ProcessIndex = 0;    
    }
  } // end while
}

//////////////////////////////////////////////////////////////////////
void sensor18b20::update()
{
  // handle only new measured and finished data
  if(m_ProcessIndex < m_MeasureIndex) 
  {
    m_ProcessIndex++;
 
     //Serial.printf("update2 %.2f\n", m_Sensor[0].m_tempCelcius);
    
    if(m_numberOfFoundSensors)
    {
      // sets m_ValidStartTime for correct Uptime Info
      if((m_ValidStartTime == "Unknown") && MyNTPClient::GetInstance().isTimeSynchronized())
      {
        char mBuffer[25];
        DateTime dt(MyNTPClient::GetInstance().getEpochTime());
        snprintf(mBuffer, sizeof(mBuffer), "%04d/%02d/%02d %02d:%02d:%02d", dt.year(), dt.month(), dt.day(), dt.hour(), dt.minute(), dt.second());

        m_ValidStartTime = mBuffer;
      }
      
      // Oled Text
      int index = 0;
      for (const auto &p : m_SensorOrderOled) 
      {
        // Kann es passieren das die Daten gleichzeitig abgefragt werden?
        m_OledData.m_LineString[index++] = &p.second->m_Name[0] + String(" ") + String(p.second->m_tempCelcius);        
        Serial.printf( " %s %.2f,", &p.second->m_Name[0], p.second->m_tempCelcius );
      }
      m_OledData.m_nLines = m_SensorOrderOled.size();
      Serial.printf("\n");

      for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
      {
        // Kann es passieren das die Daten gleichzeitig abgefragt werden?
        // write to the file
        m_SensorEntryText[thisSensor] = String(m_Sensor[thisSensor].m_tempCelcius);
      }

      char oneSensor[80]; 
      String configApp = "";

      //1;esp-simulation;0.9;2018/10/03 11:46:23;33
      //0;721E;4337;WWLVL;0;0;0;977

      for (const auto &p : m_SensorOrderWeb) 
      {
        if(configApp.length() == 0)
        {
          String DeviceName = MyBasecamp::GetInstance().configuration.get("DeviceName");
          
          // first line of config data: Interface Version;Device Network Name;Version;Uptime;MeasureCount
          snprintf(oneSensor, sizeof(oneSensor), "1;%s;%s;%s;%d\n", &DeviceName[0], VERSION_STRING, &m_ValidStartTime[0], int(p.second->m_CountGesamt));
          configApp = oneSensor;
        }
        
        // every line of config data: - Scale1(Temperature)/Scale0(kWh);Crc16Id;Temperature;Name;crc1 count;crc2 count;Measure Count
        snprintf(oneSensor, sizeof(oneSensor), "1;%s;%.2f;%s;%d;%d;%d\n", &p.second->m_crc16Name[0], p.second->m_tempCelcius, &p.second->m_Name[0], 
                                                                  int(p.second->m_CountCRC1), int(p.second->m_CountCRC2), int(p.second->m_CountSensorError));
        configApp += oneSensor;
      }

      // Kann es passieren das die Daten gleichzeitig abgefragt werden?
      m_WebAppConfigText = configApp;
    }
    else
    {
      Serial.println( "No Dallas sensors." );
    }
  }
}

