/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/
#include <arduino.h>
#include "tempLoggerESP32.h"
#include "webdebug.h"
#include "myNTPClient.h"


//////////////////////////////////////////////////////////////////////

extern TaskHandle_t OledTask;
extern TaskHandle_t TempTask;
extern TaskHandle_t _async_service_task_handle;

char *WebDebug::m_SortedRingBuffer[DEBUG_LINE_CNT];

//////////////////////////////////////////////////////////////////////
WebDebug::WebDebug()
{
  m_actIndex = 0;

  for(int i = 0; i < DEBUG_LINE_CNT; i++)
  {
    char * buf = getFreeBuffer();
    buf[0] = '\n';
    buf[1] = '\0';
  }
}

//////////////////////////////////////////////////////////////////////
WebDebug& WebDebug::GetInstance()
{
  static WebDebug instance;
  return instance;
}

//////////////////////////////////////////////////////////////////////
char* WebDebug::getFreeBuffer()
{
  if(++m_actIndex == DEBUG_LINE_CNT)
  {
    m_actIndex = 0;
  }
  DateTime dt(MyNTPClient::GetInstance().getEpochTime());
  snprintf(&m_RingBuffer[m_actIndex*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)], DEBUG_LINE_LENGTH, "%02d:%02d:%02d ", dt.hour(), dt.minute(), dt.second() );
  
  return &m_RingBuffer[m_actIndex*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)+DEBUG_TIMESTAMP-1];  
}

//////////////////////////////////////////////////////////////////////
void WebDebug::printSerial()
{
  for(int i = m_actIndex+1; i < DEBUG_LINE_CNT; i++)
  {
    Serial.printf(&m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)]);
  }
  for(int i = 0; i <= m_actIndex; i++)
  {
    Serial.printf(&m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)]);
  }
}

//////////////////////////////////////////////////////////////////////
// bring ringbuffer in sorted order
size_t WebDebug::CreateSortedRingbuffer()
{
  int act = 0;
  size_t nLen = 0;
  for(int i = m_actIndex+1; i < DEBUG_LINE_CNT; i++)
  {
    m_SortedRingBuffer[act++] = &m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)];
    nLen += strlen(&m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)])+1;
  }
  for(int i = 0; i <= m_actIndex; i++)
  {
    m_SortedRingBuffer[act++] = &m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)];
    nLen += strlen(&m_RingBuffer[i*(DEBUG_LINE_LENGTH+DEBUG_TIMESTAMP)])+1;
  }

  return nLen;
}

//////////////////////////////////////////////////////////////////////
// webserver callback function 
size_t WebDebug::printDebugData(uint8_t *buffer, size_t maxLen, size_t alreadySent)
{
  int iSent = alreadySent;

  for(int i = 0; i < DEBUG_LINE_CNT ; i++)
  {
    int lenActLine = strlen(m_SortedRingBuffer[i]);
   
    if(iSent+1 + lenActLine > maxLen)
    {
      break;
    }

    memcpy(&buffer[iSent],m_SortedRingBuffer[i], lenActLine);
    iSent += lenActLine;
    buffer[iSent++] = '\n';
  }

  if(iSent >= maxLen)
  {
    extern SemaphoreHandle_t m_SemaphoreFile;
    xSemaphoreGive(m_SemaphoreFile); // release semaphore web request
  }
  
  return maxLen; 
}

//////////////////////////////////////////////////////////////////////
void WebDebug::system_info()
{
  static TaskSnapshot_t arr[25];
  UBaseType_t sizeArr = 0;
  int nArray = uxTaskGetSnapshotAll( &arr[0], 25, &sizeArr);

  for(int i = 0; i < nArray; i++)
  {
    tskTaskControlBlock * tcb = (tskTaskControlBlock*)(arr[i].pxTCB);
    snprintf(getFreeBuffer(), DEBUG_LINE_LENGTH, "%d, %s", i+1, tcb->pcTaskName);
  }

/* 
 *  typedef enum
{
  eRunning = 0, //!< A task is querying the state of itself, so must be running. 
  eReady,     //!< The task being queried is in a read or pending ready list. 
  eBlocked,   //!< The task being queried is in the Blocked state. 
  eSuspended,   //!< The task being queried is in the Suspended state, or is in the Blocked state with an infinite time out. 
  eDeleted    //!< The task being queried has been deleted, but its TCB has not yet been freed. 
} eTaskState;
 */
 #define printTaskInfo(xTask) snprintf(getFreeBuffer(),DEBUG_LINE_LENGTH,"%s: State: %d , %d",pcTaskGetTaskName(xTask),eTaskGetState(xTask),uxTaskGetStackHighWaterMark(xTask));  

  printTaskInfo(TempTask);
  printTaskInfo(OledTask);
  printTaskInfo(_async_service_task_handle);
  
  snprintf(getFreeBuffer(), DEBUG_LINE_LENGTH, "min_free_heap: %d", esp_get_minimum_free_heap_size());
}
